/*
 *  Linux kernel adaptation layers for owl wifi library.
 *  Copyright (C) 2010 HD Wireless AB
 *
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/netdevice.h> /* dev_kfree_skb_any() */

#include <linux/mmc/host.h>
#include <linux/mmc/core.h>
#include <linux/mmc/mmc.h>
#include <linux/mmc/card.h>
#include <linux/mmc/sdio.h>
#include <linux/mmc/sdio_func.h>


#ifdef OWL_CONFIG_RAW
# include "owl_raw.h"
# include <x_test_p_6_12_17_4.h>
#else
# include <wl_fw.h>
#endif

#include "wl_api.h"

#include "owl_mmc.h"
#include "owl_drv.h"
#include "owl_net.h"

#ifdef OWL_CONFIG_LOOP
# include "owl_loop.h"
#elif defined(OWL_CONFIG_RAW)
# include "owl_raw.h"
#endif

MODULE_AUTHOR("HD Wireless AB");
MODULE_LICENSE("GPL");

/* Bit 1, Interrupt pending, write '1' to clear */
#define OWL_CCCR_INTACK     0xf0

/* Bit 0, Wakeup, read/write */
#define OWL_CCCR_WAKEUP     0xf1

/* Bit 0, Assert interrupt without SDIO CLK, read/write */
#define OWL_CCCR_NOCLK_INT  0xf2

/* Bit 0, Hi-Speed Mode Enable, read/write */
#define OWL_CCCR_HISPEED_EN 0xf3

/* Bit 0, Fifo Underrun ; Bit 1, Fifo Overrun, read/write */
#define OWL_CCCR_FIFO_STAT  0xf4

/* Bit 0, Reset all ; Bit 1, Reset all but SDIO, Active LOW */
#define OWL_CCCR_RESET      0xf5

/* Maximum sdio block size */
#define OWL_MAX_BLKSZ       512u

struct owl_drv {
        struct sdio_func *func;
        struct work_struct work;
        struct sk_buff_head tx;

        unsigned long flags;
#define IRQ_PENDING      0
#define CARD_INITIALIZED 1
#define PS_PENDING       2 /* indicates whether PS_WAKEUP has changed or not */
#define PS_WAKEUP        3 /* requested to wakeup, set by worker thread if
                            * PS_PENDING is set
                            */

        struct owl_transport transport;
};

/* Reset card. This will be performed _after_ it has been initialized
 * by the mmc/sdio stack, so it is necessary to invoke configure_card()
 * after a successful return to redo the some of the card-initialization
 * manually.
 *
 * The sdio host must be claimed before invoking this function.
 */
static int reset_card(struct sdio_func *func)
{
        int err;

        /* reset card (0x8801ea00), note that func 0 rather than
         * func->num must be addressed here.
         */
        if ((err = mmc_io_rw_direct(func->card, 1, 0,
                                    OWL_CCCR_RESET, 0x0, NULL, 1))) {
                dev_err(&func->dev, "OWL_CCCR_RESET failed, err=%d\n", err);
                return err;
        }
        dev_dbg(&func->dev, "OWL_CCCR_RESET ok\n");
        
        mdelay(20);
        return 0;
}


/* Perform card initialization after a reset or firmware download operation.
 * Basically, this function just sends cmd 3, cmd 7 and then cmd 52 a couple of
 * times, but with a lot of traces.
 *
 * The sdio host must be claimed before invoking this function.
 */
static int configure_card(struct sdio_func *func)
{
        int err;
        struct mmc_card *card = func->card;
        struct mmc_host *host = card->host;
        
        /* cmd 3 */
        if ((err = mmc_send_relative_addr(host, &card->rca))) {
                dev_err(&func->dev, "MMC_SET_RELATIVE_ADDR, err=%d\n", err);
                return err;
        }
        dev_dbg(&func->dev, "MMC_SET_RELATIVE_ADDR ok, rca=%d\n", card->rca);

        /* cmd 7 */
        if ((err = mmc_select_card(card))) {
                dev_err(&func->dev, "MMC_SELECT_CARD, err=%d", err);
                return err;
        } 
        dev_dbg(&func->dev, "MMC_SELECT_CARD ok, rca=%d\n", card->rca);

        /* if host supports 4-bit mode, then  enable it since it will be
         * assumed by the host
         */
        if (host->caps & MMC_CAP_4_BIT_DATA) {
                
                /* enable sdio 4-bit mode (0x88000e02) */
                if ((err = mmc_io_rw_direct(card, 1, 0, SDIO_CCCR_IF,
                                            SDIO_BUS_WIDTH_4BIT, NULL, 1))) {
                        dev_err(&func->dev, "SDIO_CCCR_IF, err=%d", err);
                        return err;
                }
                dev_dbg(&func->dev, "SDIO_CCCR_IF configured\n");

        } else {
                dev_dbg(&func->dev, "MMC_CAP_4_BIT_DATA not supported by host, "
                        "host->caps=0x%08lx\n", host->caps);
        }

        /* if host supports HS mode, then  enable it since it will be
         * assumed by the host
         */
        if (host->caps & MMC_CAP_SD_HIGHSPEED) {
                
                /* enable sdio HS mode */
                if ((err = mmc_io_rw_direct(card, 1, 0, OWL_CCCR_HISPEED_EN,
                                            0x1, NULL, 1))) {
                        dev_err(&func->dev, "OWL_CCCR_HISPEED_EN, err=%d", err);
                        return err;
                }
                dev_dbg(&func->dev, "OWL_CCCR_HISPEED_EN configured\n");

        } else {
                dev_dbg(&func->dev, "MMC_CAP_SD_HIGHSPEED not supported by "
                        "host, host->caps=0x%08lx\n", host->caps);
        }

        /* if host supports SDIO IRQ, then  enable it since it will be
         * assumed by the host
         */
        if (host->caps & MMC_CAP_SDIO_IRQ) {

                /* set master int mask and enable interrupt for function
                   (0x88000803) */
                if ((err = mmc_io_rw_direct(card, 1, 0, SDIO_CCCR_IENx,
                                            0x1 | (1 << func->num), NULL, 1))) {
                        dev_err(&func->dev, "SDIO_CCCR_IENx func=%d, err=%d\n",
                                func->num, err);
                        return err;
                }
                dev_dbg(&func->dev, "SDIO_CCCR_IENx set for func %d\n",
                        func->num);

                /* allow int to be asserted when there's no sdio clk
                   (0x8801e401) */
                if ((err = mmc_io_rw_direct(card, 1, 0, OWL_CCCR_NOCLK_INT, 0x1,
                                            NULL, 1))) {
                        dev_err(&func->dev, "OWL_CCCR_NOCLK_INT, err=%d\n",
                                err);
                        return err;
                }
                dev_dbg(&func->dev, "OWL_CCCR_NOCLK_INT configured\n");
                
        } else {
                dev_dbg(&func->dev, "MMC_CAP_SDIO_IRQ not supported by "
                        "host, host->caps=0x%08lx\n", host->caps);
        }

        return 0;
}


static void pad_skb(struct owl_drv *priv, struct sk_buff *skb)
{
        struct sdio_func *func = priv->func;
        size_t n;

        /* just inform us about unaligned stuff */
        if ((unsigned int) skb->data % 4)
                dev_warn(&func->dev, "unaligned skb data=%p\n", skb->data);

        if (skb->len % 4 == 0)
                return;

        dev_dbg(&func->dev, "fixup skb len=%d\n", skb->len); 

        n = 4 - skb->len % 4;
#ifdef OWL_CONFIG_FW_LEGACY
        ((uint8_t *) skb->data)[5] = n;
#else        
        (*(uint16_t *) (skb->data + 6)) = cpu_to_le16(n);
#endif        
        if((size_t) skb_tailroom(skb) >= n) {
                memset(skb_put(skb, n), 0, n);
        } else {
                dev_warn(&func->dev, "not enough tailroom to pad skb\n");
                skb->len += n;
        }

        (*(uint16_t *) skb->data) = cpu_to_le16(skb->len - 2);
}


static void owl_drv_work(struct work_struct *work)
{
        int err;
        struct sk_buff *skb;
        struct owl_drv *priv =
                container_of(work, struct owl_drv, work);
        struct owl_transport *transport = &priv->transport;
        struct sdio_func *func = priv->func;

        sdio_claim_host(func);
        
        if (test_and_clear_bit(PS_PENDING, &priv->flags)) {
                u8 wakeup = test_bit(PS_WAKEUP, &priv->flags);
                if ((err = mmc_io_rw_direct(func->card, 1, 0, OWL_CCCR_WAKEUP,
                                            wakeup, NULL, 1)))
                        dev_err(&func->dev, "OWL_CCCR_WAKEUP (%d), err=%d\n",
                                wakeup, err);
                else
                        dev_dbg(&func->dev, "OWL_CCCR_WAKEUP: %d\n", wakeup);
        }
        
        if (test_and_clear_bit(IRQ_PENDING, &priv->flags)) {

                struct sk_buff *skb1, *skb2;
                int len;
                
                if ((skb1 = dev_alloc_skb(OWL_MIN_HICSZ)) == NULL) {
                        dev_err(&func->dev, "out of memory (%d bytes)\n",
                                OWL_MIN_HICSZ);
                        goto tx_queue;
                }
                
                if ((err = sdio_readsb(func, skb1->data, 0,
                                       OWL_MIN_HICSZ))) {
                        dev_err(&func->dev, "readsb() failed, err=%d\n", err);
                        dev_kfree_skb(skb1);
                        goto tx_queue;
                }
                len = le16_to_cpu(*((unsigned short *) skb1->data)) + 2;

                if (len > OWL_MAX_HICSZ) {
                        dev_err(&func->dev, "Invalid length %u", len);
                        goto tx_queue;
                } else if (len <= OWL_MIN_HICSZ) {
                        skb_put(skb1, len);
                } else {
                        /* more bytes */

                        if ((skb2 = dev_alloc_skb(len)) == NULL) {
                                dev_err(&func->dev,
                                        "out of memory (%u bytes)", len);
                                dev_kfree_skb(skb1);
                                goto tx_queue;
                        }
                
                        memcpy(skb_put(skb2, OWL_MIN_HICSZ), skb1->data,
                               OWL_MIN_HICSZ);
                        dev_kfree_skb(skb1);
                        len -= OWL_MIN_HICSZ;
                        err = sdio_readsb(func, skb_put(skb2, len), 0, len);
                        if (err) {
                                dev_err(&func->dev,
                                        "readsb() failed, err %d", err);
                                dev_kfree_skb(skb2);
                                goto tx_queue;
                        }
                        skb1 = skb2;
                }
#ifdef VERBOSE_DEBUG
                print_hex_dump_bytes("owl RX ", DUMP_PREFIX_NONE,
                                     skb1->data, skb1->len);
#endif

#ifdef OWL_CONFIG_LOOP
                owl_loop_rx(transport, skb1);
#elif defined(OWL_CONFIG_RAW)
                owl_raw_rx(transport, skb1);
#else
                owl_net_rx(transport, skb1);
#endif
        }

tx_queue:
        /* The list lock is taken by skb_dequeue() so the function
         * may be used safely with other locking list functions.
         */
        while ((skb = skb_dequeue(&priv->tx))) {
#ifdef VERBOSE_DEBUG
                print_hex_dump_bytes("owl TX ", DUMP_PREFIX_NONE,
                                     skb->data, skb->len);
#endif
                pad_skb(priv, skb);
                if ((err = sdio_writesb(func, 0, skb->data, skb->len))) {
                        dev_err(&func->dev, "write of %p (%d) failed, err=%d\n",
                                skb, skb->len, err);
                }
                dev_kfree_skb(skb);
        }

        sdio_release_host(func);
}


/*
 * Invoked when an SDIO interrupt occurs. The host is always claimed already
 * when the handler is called.
 *
 */
static void owl_drv_irq(struct sdio_func *func)
{
        int err;
        struct owl_drv *priv = sdio_get_drvdata(func);

        dev_dbg(&func->dev, "rx irq\n");
        if ((err = mmc_io_rw_direct(func->card, 1, 0,
                                    OWL_CCCR_INTACK, 0x02, NULL, 1))) {
                dev_err(&func->dev, "failed to ack int, err=%d\n", err);
                return;
        }
        
        /* since the irq is claimed during driver probe, we might get
         * false interrupts before the card is reset and initialized.
         */
        if (!test_bit(CARD_INITIALIZED, &priv->flags)) {
                dev_dbg(&func->dev, "ignoring spurios interrupt before card "
                        "was initialized\n");
                return;
        }
        
        /* should never happen */
        if (test_bit(IRQ_PENDING, &priv->flags))
                dev_warn(&func->dev, "got new irq before rx was completed\n");
        
        set_bit(IRQ_PENDING, &priv->flags);
        schedule_work(&priv->work);
}

static size_t fw_read_cb(void *ctx, const uint8_t **buf,
                         size_t offset, size_t len)
{
        if (NULL == buf)
                return 0;
        
        *buf = ((uint8_t *) fw_buf) + offset;
        if (len > (fw_len - offset))
                return fw_len - offset;

        return len;
}


/*
 * Invoked from wl_api when the a management message or a firmware chunk should
 * be transmitted to the WiFi device. This function should ensure that the
 * message is passed through to the device and should never fail.
 *
 * This function used to live in owl_net.c, but since we now also handle
 * firmware download through mgmt_tx_cb() it was moved here so we can test
 * the CARD_INITIALIZED bit.
 *
 */
static void mgmt_tx_cb(void *ctx, const uint8_t *buf, uint16_t len)
{
        struct owl_drv *priv = ctx;
        struct sdio_func *func = priv->func;
        int err;

        static int foo = 0;

        if (foo)
            return;
        
        if (!test_bit(CARD_INITIALIZED, &priv->flags)) {
                /* this must be a firmware chunk, so call the sdio layer
                   directly since more chunks will follow without yield */
                
                unsigned char *dma_buf;
                if ((dma_buf = kmalloc(len, GFP_KERNEL | GFP_DMA)) == NULL) {
	                dev_err(&func->dev, "kmalloc of %d fail\n", len); 
	                return;
	        }
                memcpy(dma_buf, buf, len);

                sdio_claim_host(func);
                if ((err = sdio_writesb(func, 0, dma_buf, len))) {
                        dev_err(&func->dev,
                                "sdio_writesb fail, err:%d len:%d\n", err, len);
                        foo = 1;
                }
                sdio_release_host(func);
                kfree(dma_buf);
                
        } else {
                /* mgmt frame */
                
                struct sk_buff *skb;
                if ((skb = dev_alloc_skb(len)) == NULL) {
                        dev_err(&func->dev,
                                "fail to alloc skb for mgmt len:%d\n", len);
                        return;
                }
                memcpy(skb_put(skb, len), buf, len);
                priv->transport.tx(&priv->transport, skb);
        }
        
}


/*
 * Should be invoked when interface is about to be brought up
 *
 */
static int owl_drv_fw_download(struct owl_transport *transport)
{
        int err;

        struct owl_drv *priv =
                container_of(transport, struct owl_drv, transport);
        struct sdio_func *func = priv->func; 
        
        sdio_claim_host(func);

        err = reset_card(func);
        if (err)
                goto err_release;

        err = configure_card(func);
        if (err)
                goto err_release;

        sdio_release_host(func);

        dev_dbg(&func->dev, "fw_download start\n");

        /* must be invoked before wl_init() and fw_download */
        wl_register_mgmt_tx_cb(mgmt_tx_cb, priv);

        if ((err = wl_fw_download(fw_read_cb, priv)) != WL_SUCCESS)
                return -EFAULT;

        dev_dbg(&func->dev, "fw_download done, wait for card to start\n");
        mdelay(20);

        sdio_claim_host(func);

        err = configure_card(func);
        if (err)
                goto err_release;

        /* since the irq is claimed during driver probe, we might get
         * false interrupts before the card is reset and initialized.
         * this flag will tell the irq handler that all interrupts from now
         * on are 'real'
         */
        set_bit(CARD_INITIALIZED, &priv->flags);

err_release:
        sdio_release_host(func);
        return err;
}


/*
 * Send a hic message to the card
 *
 */
static void owl_drv_tx(struct owl_transport *transport, struct sk_buff *skb)
{
        struct owl_drv *priv =
                container_of(transport, struct owl_drv, transport);
        
        /* skb_queue_tail takes the list lock and can be used safely with other
         * locking sk_buff functions
         */
        skb_queue_tail(&priv->tx, skb);
        schedule_work(&priv->work); 
}


/*
 * Set wakeup bit to enable/disable power save mode on card.
 *
 */
static void owl_drv_wakeup(struct owl_transport *transport, int wakeup)
{
        struct owl_drv *priv =
                container_of(transport, struct owl_drv, transport);

        /* just inform the worker thread to enable ps */
        if (wakeup)
                set_bit(PS_WAKEUP, &priv->flags);
        else
                clear_bit(PS_WAKEUP, &priv->flags);
        
        set_bit(PS_PENDING, &priv->flags);
        schedule_work(&priv->work);
}


/*
 * Invoked when any card is detected in SDIO slot. Should initialize the card
 * and the wl_api.
 */
static int owl_drv_probe(struct sdio_func *func,
                         const struct sdio_device_id *id)
{
        struct owl_drv *priv;
        struct mmc_card *card = func->card;
        int err;

        dev_dbg(&func->dev, "probe class=%u vendor=%04x device=%04x\n",
                func->class, func->vendor, func->device);

        if (func->vendor != 0x03bb) {
                dev_info(&func->dev, "unknown vendor (0x%04x)\n", func->vendor);
                return -ENOSYS;
        }
        
        dev_dbg(&func->dev, "card->cccr reported high_speed=%d wide_bus=%d "
                "high_power=%d low_speed=%d multi_block=%d\n",
                card->cccr.high_speed, card->cccr.wide_bus,
                card->cccr.high_power, card->cccr.low_speed,
                card->cccr.multi_block);

        dev_dbg(&func->dev, "card->cis reported max_dtr=%d\n",
                card->cis.max_dtr);

        dev_dbg(&func->dev, "host->caps are 0x%08lx\n", card->host->caps);
        
        priv = kzalloc(sizeof(*priv), GFP_KERNEL);
        if (!priv) {
                dev_warn(&func->dev, "out of memor\n");
                return -ENOMEM;
        }

        priv->func = func;
        sdio_set_drvdata(func, priv);
        INIT_WORK(&priv->work, owl_drv_work);
        skb_queue_head_init(&priv->tx);
        
        priv->transport.tx = owl_drv_tx;
        priv->transport.fw_download = owl_drv_fw_download;
        priv->transport.wakeup = owl_drv_wakeup;
        
        sdio_claim_host(func);

        err = sdio_enable_func(func);
        if (err) {
                dev_err(&func->dev, "failed to enable func, err = %d", err);
                goto err;
        }
        dev_dbg(&func->dev, "function %d enabled\n", func->num);

        /* Card CCCR reports multi_block:1 high_speed:0 wide_bus:0 falsely,
         * fixup these values here.
         */
        card->cccr.multi_block = 0;
        card->cccr.high_speed = 1;
        card->cccr.wide_bus = 1;
        dev_dbg(&func->dev,
                "fixup cccr, multi_block:%d high_speed:%d wide_bus:%d\n",
                card->cccr.multi_block, card->cccr.high_speed,
                card->cccr.wide_bus);
        
        /* the card reports a block size of 16 even though it supports 512
         * byte blocks.
         */
        if (func->max_blksize != OWL_MAX_BLKSZ) {
                unsigned int blksz =
                        min(OWL_MAX_BLKSZ, card->host->max_blk_size);

                dev_dbg(&func->dev, "reported block size was %d, set to %d\n",
                        func->max_blksize, blksz);
                
                err = sdio_set_block_size(func, blksz);
                if (err) {
                        dev_err(&func->dev, "failed to set block size, err=%d",
                                err);
                        sdio_disable_func(func);
                        goto err;
                }
 
                /* in drivers/mmc/core/sdio_io.c, sdio_set_block_size()
                 * updates func->cur_blksize which is used for block transfers.
                 * func->max_blksize is unchanged, but still used by
                 * sdio_max_byte_size() for byte transfers. Therefore, we'll
                 * force update this field here.
                 */
                func->max_blksize = OWL_MAX_BLKSZ;
        }

        err = sdio_claim_irq(func, owl_drv_irq);
        if (err) {
                dev_err(&func->dev, "failed to claim irq, err %d\n", err);
                sdio_disable_func(func);
                goto err;
        }
        dev_dbg(&func->dev, "irq claimed\n");

        /* release host here since the call to owl_net_register_drv()
         * might try to download firmare and stuff. Make that anything that
         * goes wrong beyond this point has claimed the bus again since it will
         * be released in the err label.
         */
        sdio_release_host(func);
        
        
#ifdef OWL_CONFIG_LOOP
        err = owl_loop_register_drv(&priv->transport);
#elif defined(OWL_CONFIG_RAW)
        err = owl_raw_register_drv(&priv->transport);
#else
        err = owl_net_register_drv(&priv->transport);
#endif
        if (err) {
                sdio_claim_host(func);
                sdio_release_irq(func);
                sdio_disable_func(func);
                dev_err(&func->dev, "failed register drv, err=%d", err);
                goto err; /* host will be released in err label */
        }

        return 0;

err:
        sdio_release_host(func);
        kfree(priv);
        return err;
}


/*
 * Invoked by the mmc/sdio stack when the card is removed from the sdio slot.
 * Should cleanup any resources used by wl_api.
 *
 */
static void owl_drv_remove(struct sdio_func *func)
{
        struct sk_buff *skb;
        struct owl_drv *priv = sdio_get_drvdata(func);

        dev_dbg(&func->dev, "remove\n");

        /* remove will never be called unless probe succeeded */

#ifdef OWL_CONFIG_LOOP
        /* nop */
#elif defined(OWL_CONFIG_RAW)
        owl_raw_unregister_drv(&priv->transport);
#else
        owl_net_unregister_drv(&priv->transport);
#endif

        flush_scheduled_work();

        while ((skb = skb_dequeue(&priv->tx)))
                dev_kfree_skb_any(skb); /* if in irq context */

        sdio_claim_host(func);
        sdio_release_irq(func);
        sdio_disable_func(func);
        sdio_release_host(func);
}

static const struct sdio_device_id owl_drv_ids[] = {
	{ SDIO_DEVICE(0x03bb, 0) },
	{ SDIO_DEVICE(0x03bb, 1) },
	{ /* end: all zeroes */ }
};

MODULE_DEVICE_TABLE(sdio, owl_drv_ids);

static struct sdio_driver owl_sdio_driver = {
	.probe		= owl_drv_probe,
	.remove		= owl_drv_remove,
	.name		= "owl_sdio",
	.id_table	= owl_drv_ids,
};

static int __init owl_drv_init(void)
{
        return sdio_register_driver(&owl_sdio_driver);
}

static void __exit owl_drv_exit(void)
{
        sdio_unregister_driver(&owl_sdio_driver);
}

module_init(owl_drv_init);
module_exit(owl_drv_exit);
