/*
 *  Linux kernel adaptation layers for owl wifi library.
 *  Copyright (C) 2010 HD Wireless AB
 *
 */

#include <linux/sched.h>

#include "owl_sync.h"
#include "owl_wext_event.h"

#ifdef OWL_CONFIG_FW_AP
#include <wlap_api.h>
#endif

static void init_complete_cb(void *ctx)
{
        struct owl_net *priv = ctx;
        struct net_device *dev = priv->dev;
        
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);
        set_bit(INIT_COMPLETE, &priv->flags);
        wake_up_interruptible(&priv->waitq);
}

#ifdef OWL_CONFIG_FW_AP
static void _event_mic_failure(struct wl_mac_addr_t src, 
                               int unicast, 
                               void *ctx) {

}

/* Receives a raw ethernet frame */
static void _event_eapol_frame(const char *pkt,
                               const size_t pkt_len,
                               void *ctx) {
}

static void _event_sta_assoc(const enum wlap_assoc_event_t event,
                             const struct wl_mac_addr_t addr,
                             const int reason,
                             const char *ies,
                             const size_t ielen,
                             void *ctx) {
        struct owl_net *priv = ctx;
        struct net_device *dev = priv->dev;

        dev_dbg(&dev->dev, "ENTER %s: event %d\n", __func__, (int)event);
        switch (event) {
        case WL_EVENT_STA_DISASSOC:
                break;
        case WL_EVENT_STA_DEAUTH:
                break;
        case WL_EVENT_STA_ASSOC:
                break;
                
        default:
                break;
        }
}

#else /* OWL_CONFIG_FW_AP */
static void mic_failure_cb(void *ctx)
{
        struct owl_net *priv = ctx;
        struct net_device *dev = priv->dev;
        
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);
        owl_wext_event_micfailure(priv);
}

#endif /* OWL_CONFIG_FW_AP */

/* This function will be invoked in the context of the call to wl_process_rx().
 * If owl_sync_wl_process_rx() is used, the wl_sema will be held during the
 * execution of this function.
 */
static void event_cb(struct wl_event_t event, void* ctx)
{
        struct owl_net *priv = ctx;
        struct net_device *dev = priv->dev; 

        dev_dbg(&dev->dev, "ENTER %s\n", __func__);

        switch (event.id) {
        case WL_EVENT_MEDIA_CONNECTED: {
                struct wl_network_t *net;
                dev_info(&dev->dev, "%s media connected\n", __func__);
                if ((net = wl_get_current_network()) && wl_is_connected())
                        owl_wext_event_ap(priv, &net->bssid);
                else
                        dev_err(&dev->dev, "failed to get current net\n");

                netif_carrier_on(dev);
                break;
        }
                
        case WL_EVENT_CONN_FAILURE:
                dev_info(&dev->dev, "%s conn failure\n", __func__);
                break;
                
        case WL_EVENT_MEDIA_DISCONNECTED:
                dev_info(&dev->dev, "%s media disconnected\n", __func__);
                owl_wext_event_ap(priv, NULL);
                netif_carrier_off(dev);

                /* owl_sync_wl_connect() might be waiting for this to happen */
                wake_up_interruptible(&priv->waitq);
                break;

        case WL_EVENT_SCAN_COMPLETE:
                dev_dbg(&dev->dev, "%s scan complete\n", __func__);
                owl_wext_event_scan_complete(priv);

                set_bit(SCAN_COMPLETE, &priv->flags);
                wake_up_interruptible(&priv->waitq);
                break;
                
        default:
                dev_err(&dev->dev, "%s unhandled event %d\n", __func__,
                        event.id);
                break;
        };
}

int owl_sync_wl_init(struct owl_net *priv)
{
        wl_err_t wl_err;
        int err;
        struct net_device *dev = priv->dev;
        
        clear_bit(INIT_COMPLETE, &priv->flags);
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return err;
        }
        if ((wl_err = wl_init(priv, init_complete_cb,
                              WL_HOST_ATTENTION_SDIO)) != WL_SUCCESS) {
                up(&priv->wl_sema);
                dev_err(&dev->dev, "init failed, err=%d\n", wl_err);
                return -EIO;
        }
#ifdef OWL_CONFIG_FW_AP
        wlap_register_mic_failure_cb(_event_mic_failure, priv);
        wlap_register_eapol_frame_rx_cb(_event_eapol_frame, priv);
        wlap_register_assoc_event_cb(_event_sta_assoc, priv);
#else /* OWL_CONFIG_FW_AP */
        (void) wl_register_mic_failure_cb(mic_failure_cb, priv);
#endif /* OWL_CONFIG_FW_AP */
        (void) wl_register_event_cb(event_cb, priv);
        up(&priv->wl_sema);

        if ((err = wait_event_interruptible_timeout(
                     priv->waitq,
                     test_bit(INIT_COMPLETE, &priv->flags), 10 * HZ)) <= 0) {
                
                dev_err(&dev->dev, "wl_init did not complete, err=%d\n", err);
                return err;
        }

        return 0;
}

void owl_sync_wl_shutdown(struct owl_net *priv)
{
        wl_err_t wl_err;
        int err;
        struct net_device *dev = priv->dev;
        
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return;
        }
        wl_err = wl_shutdown();
        up(&priv->wl_sema);

        if (wl_err != WL_SUCCESS) {
                dev_err(&dev->dev, "shutdown failed, err=%d\n", wl_err);
                return;
        }

        /* dont call owl_wext_event_ap(priv, NULL) at this time, since that
           will trigger e.g. the wpa_supplicat to reconnect while the driver
           is shutting down */
        
        netif_carrier_off(dev);
}

wl_err_t owl_sync_wl_process_rx(struct owl_net *priv,
                            char *pkt, size_t pkt_len, char **stripped_pkt, 
                            size_t *stripped_pkt_len, uint16_t *vlanid_prio)
{
        wl_err_t wl_err;
        int err;
        struct net_device *dev = priv->dev;

        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return WL_RESOURCES; /* retry */
        }
        wl_err = wl_process_rx(pkt, pkt_len, stripped_pkt, stripped_pkt_len,
                               vlanid_prio);
        up(&priv->wl_sema);
        return wl_err;
}

wl_err_t owl_sync_wl_process_tx(struct owl_net *priv,
                                char *eth_hdr, 
                                size_t eth_hdr_len,
				size_t pkt_len, 
				char *hdr, 
                                uint16_t vlanid_prio,
                                void *pkt_handle)
{
        wl_err_t wl_err;
        struct net_device *dev = priv->dev;

        if (down_trylock(&priv->wl_sema)) {
                dev_dbg(&dev->dev, "%s sema busy\n", __func__);
                return WL_RETRY; /* retry pkt later */
        }
        wl_err = wl_process_tx(eth_hdr, 
                               eth_hdr_len, 
                               pkt_len, 
                               hdr, 
                               vlanid_prio,
                               pkt_handle);
        up(&priv->wl_sema);
        return wl_err;
}

wl_err_t owl_sync_wl_scan(struct owl_net *priv, struct wl_ssid_t *ssid)
{
        wl_err_t wl_err;
        int err;
        struct net_device *dev = priv->dev;

        clear_bit(SCAN_COMPLETE, &priv->flags);
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return WL_FAILURE;
        }
        wl_err = wl_scan(ssid);
        up(&priv->wl_sema);

        if (wl_err != WL_SUCCESS)
                return wl_err;
        
        err = wait_event_interruptible_timeout(
                priv->waitq,
                test_bit(SCAN_COMPLETE, &priv->flags), 2 * HZ);
        
        if (err <= 0) {
                dev_err(&dev->dev, "scan did not complete, err=%d\n", err);
                return WL_FAILURE;
        }
        
        return wl_err;
}

/* return int rather than wl_network_t ptr to be able to indicate -EINTR to
 * caller
 *
 * @return -EINTR if interrupted by a signal
 *         
 */
int owl_sync_wl_get_current_network(struct owl_net *priv,
                                    struct wl_network_t **net)
{
        int err;
        struct net_device *dev = priv->dev;
        struct wl_network_t *__net;

        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return err;
        }
        __net = wl_get_current_network();
        *net = NULL;
        if (__net) {
                *net = kmalloc(sizeof(*__net), GFP_KERNEL);
                if (*net == NULL) {
                        up(&priv->wl_sema);
                        return -ENOMEM;
                }
                memcpy(*net, __net, sizeof(*__net));
        }
        
        up(&priv->wl_sema);
        return 0;
}

int owl_sync_wl_get_network_list(struct owl_net *priv,
                                 struct wl_network_list_t* netlist)
{
        int err;
        wl_err_t wl_err;
        int i;
        struct net_device *dev = priv->dev;
        struct wl_network_list_t *ilist;

        BUG_ON(!netlist);
                
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return err;
        }

        if ((wl_err = wl_get_network_list(&ilist)) != WL_SUCCESS) {
                up(&priv->wl_sema);
                dev_warn(&dev->dev, "failed to get network list\n");
                return -EIO;
        }

        netlist->net = kzalloc(sizeof(struct wl_network_t *) *
                               ilist->cnt, GFP_KERNEL);
        if (netlist->net == NULL) {
                up(&priv->wl_sema);
                dev_warn(&dev->dev, "failed to get alloc net array\n");
                return -ENOMEM;
        }

        netlist->cnt = 0;
        for (i = 0; i < ilist->cnt; i++) {
                struct wl_network_t *net = ilist->net[i];
                netlist->net[i] =
                        kmalloc(sizeof(*net) + net->ie_len, GFP_KERNEL);
                
                if (netlist->net[i] == NULL) {
                        up(&priv->wl_sema);
                        dev_warn(&dev->dev, "could not allocate space for all "
                                 "networks\n");
                        owl_sync_put_network_list(netlist);
                        return -ENOMEM;
                }
                
                memcpy(netlist->net[i], net, sizeof(*net) + net->ie_len);
                netlist->cnt++;
        }

        up(&priv->wl_sema);
        return 0;
}

void owl_sync_put_network_list(struct wl_network_list_t *netlist)
{
        int i;
        BUG_ON(netlist == NULL);

        for (i = 0; i < netlist->cnt; i++)
                kfree(netlist->net[i]);

        kfree(netlist->net);
}


/*
 * To be nice to wireless tools, we'll disconnect from the current net if
 * trying to connect to another ap without doing 'iwconfig owl0 essid off'
 * first. If the new ap is the same as the old one, we'll reconnect.
 *
 */
wl_err_t owl_sync_wl_connect(struct owl_net *priv, char* ssid, size_t ssid_len)
{
        int err;
        wl_err_t wl_err;
        struct net_device *dev = priv->dev;
        
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return WL_FAILURE;
        }

        /* if we are currently connected to a net (same or different ssid),
         * disconnect from that network first, otherwise wl_connect will return
         * WL_BUSY
         */
        if (wl_is_connected()) {
                dev_info(&dev->dev, "%s disconnecting first\n", __func__);
                if ((wl_err = wl_disconnect()) != WL_SUCCESS)
                        goto out;
                
                /* release semaphore so we don't deadlock the events */
                up(&priv->wl_sema);
                
                /* block here until we're disconnected */
                if ((err = wait_event_interruptible_timeout(
                             priv->waitq, !netif_carrier_ok(dev),
                             5 * HZ)) < 0) {
                        dev_err(&dev->dev, "%s wait for carrier off failed, "
                                "err=%d\n", __func__, err);
                        return WL_FAILURE;
                }
                
                /* re-aquire semaphore */
                if ((err = down_interruptible(&priv->wl_sema))) {
                        dev_info(&dev->dev, "%s interrupted, err=%d\n",
                                 __func__, err);
                        return WL_FAILURE;
                }
        }
        wl_err = wl_connect(ssid, ssid_len);
        
out:
        up(&priv->wl_sema);
        return wl_err;
}

wl_err_t owl_sync_wl_disconnect(struct owl_net *priv)
{
        int err;
        wl_err_t wl_err;
        struct net_device *dev = priv->dev;
        
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return WL_FAILURE;
        }
        wl_err = wl_disconnect();
        up(&priv->wl_sema);

        return wl_err;
}

/* similar, but not the same, as owl_sync_wl_connect() */
wl_err_t owl_sync_wl_connect_bssid(struct owl_net *priv,
                                   struct wl_mac_addr_t bssid)
{
        int err;
        wl_err_t wl_err;
        struct net_device *dev = priv->dev;
        struct wl_network_t *net;
            
        
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return WL_FAILURE;
        }

        /* if we are currently connected to a net (same or different ssid),
         * disconnect from that network first, otherwise wl_connect will return
         * WL_BUSY
         */
        if ((wl_is_connected() && (net = wl_get_current_network()))) {

                /* trying to connect to same net? this want cause a reconnect
                 * as in the owl_sync_wl_connect()-case, see owl_wext_set_ap()
                 * for more info
                 */
                if (memcmp(net->bssid.octet, bssid.octet,
                           WL_MAC_ADDR_LENGTH) == 0) {
                        dev_info(&dev->dev, "%s already connected to net\n",
                                 __func__);
                        wl_err = WL_SUCCESS;
                        goto out;
                }
                                
                dev_info(&dev->dev, "%s disconnecting first\n", __func__);
                if ((wl_err = wl_disconnect()) != WL_SUCCESS)
                        goto out;
                
                /* release semaphore so we don't deadlock the events */
                up(&priv->wl_sema);
                
                /* block here until we're disconnected */
                if ((err = wait_event_interruptible_timeout(
                             priv->waitq, !netif_carrier_ok(dev),
                             5 * HZ)) < 0) {
                        dev_err(&dev->dev, "%s wait for carrier off failed, "
                                "err=%d\n", __func__, err);
                        return WL_FAILURE;
                }
                
                /* re-aquire semaphore */
                if ((err = down_interruptible(&priv->wl_sema))) {
                        dev_err(&dev->dev, "%s interrupted, err=%d\n",
                                __func__, err);
                        return WL_FAILURE;
                }
        }
        wl_err = wl_connect_bssid(bssid);
        
out:
        up(&priv->wl_sema);
        return wl_err;
}

wl_err_t owl_sync_wl_get_rate(struct owl_net *priv,
                              wl_rate_t *tx, wl_rate_t *rx)
{
        int err;
        wl_err_t wl_err;
        struct net_device *dev = priv->dev;
        
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return WL_FAILURE;
        }
        wl_err = wl_get_rate(tx, rx);
        up(&priv->wl_sema);

        return wl_err;        
}

wl_err_t owl_sync_wl_set_key(struct owl_net *priv,
                             uint8_t key_idx, 
                             size_t key_len,
                             const void *key,
                             int swap,
                             const struct wl_mac_addr_t *bssid,
                             const struct wl_receive_seq_cnt_t *rsc,
                             int tx_key)
{
#ifdef OWL_CONFIG_FW_AP
        return WL_NOT_SUPPORTED;
#else /* OWL_CONFIG_FW_AP */
        int err;
        wl_err_t wl_err;
        struct net_device *dev = priv->dev;
        
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return WL_FAILURE;
        }
        wl_err = wl_set_key(key_idx, key_len, key, swap, bssid, rsc, tx_key);
        up(&priv->wl_sema);

        return wl_err;
#endif /* OWL_CONFIG_FW_AP */
}

wl_err_t owl_sync_wl_add_wep_key(struct owl_net *priv,
                                 uint8_t key_idx, 
                                 size_t key_len, 
                                 const void *key, 
                                 struct wl_mac_addr_t *bssid)
{
        int err;
        wl_err_t wl_err;
        struct net_device *dev = priv->dev;
        
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return WL_FAILURE;
        }
        wl_err = wl_add_wep_key(key_idx, key_len, key, bssid);
        up(&priv->wl_sema);

        return wl_err;
}

wl_err_t owl_sync_wl_set_default_wep_key(struct owl_net *priv,
                                         uint8_t key_idx)
{
        int err;
        wl_err_t wl_err;
        struct net_device *dev = priv->dev;
        
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return WL_FAILURE;
        }
        wl_err = wl_set_default_wep_key(key_idx);
        up(&priv->wl_sema);

        return wl_err;
}

wl_err_t owl_sync_wl_delete_wep_key(struct owl_net *priv,
                                    uint8_t key_idx,
                                    struct wl_mac_addr_t *bssid)
{
        int err;
        wl_err_t wl_err;
        struct net_device *dev = priv->dev;
        
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return WL_FAILURE;
        }
        wl_err = wl_delete_wep_key(key_idx, bssid);
        up(&priv->wl_sema);

        return wl_err;
}

wl_err_t owl_sync_wl_enable_ps(struct owl_net *priv)
{
        int err;
        wl_err_t wl_err;
        struct net_device *dev = priv->dev;
        
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return WL_FAILURE;
        }
        wl_err = wl_enable_ps();
        up(&priv->wl_sema);

        return wl_err;
}

wl_err_t owl_sync_wl_disable_ps(struct owl_net *priv)
{
        int err;
        wl_err_t wl_err;
        struct net_device *dev = priv->dev;
        
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return WL_FAILURE;
        }
        wl_err = wl_disable_ps();
        up(&priv->wl_sema);

        return wl_err;
}

wl_err_t owl_sync_wl_conf_ps(struct owl_net *priv,
                             uint8_t use_ps_poll, uint32_t ps_traffic_timeout,
                             uint32_t ps_delay, uint8_t rx_all_dtim,
                             uint16_t listen_interval)
{
        int err;
        wl_err_t wl_err;
        struct net_device *dev = priv->dev;
        
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return WL_FAILURE;
        }
        wl_err = wl_conf_ps(use_ps_poll, ps_traffic_timeout, ps_delay,
                            rx_all_dtim, listen_interval);
        up(&priv->wl_sema);
        return wl_err;    
}

void owl_sync_wl_poll(struct owl_net *priv, uint32_t tick)
{
        int err;
        struct net_device *dev = priv->dev;
        
        if ((err = down_interruptible(&priv->wl_sema))) {
                dev_info(&dev->dev, "%s interrupted, err=%d\n", __func__, err);
                return;
        }
        wl_tick(tick);
        up(&priv->wl_sema);
}

#ifdef OWL_CONFIG_FW_AP

int owl_sync_get_mode()
{
        return IW_MODE_MASTER;
}

wl_err_t owl_sync_wlap_start(struct owl_net *priv,
                             char *ssid,
                             int ssid_len)
{
        return wlap_start_ap(ssid, ssid_len, priv->channel);
}

wl_err_t owl_sync_set_channel(struct owl_net *priv,
                              uint8_t channel) {
        priv->channel = channel;
        return WL_SUCCESS;
}

wl_err_t owl_sync_wl_set_ssid(struct owl_net *priv, char* ssid, size_t ssid_len)
{
        struct net_device *dev = priv->dev;
        
        dev_info(&dev->dev, "%s called\n", __func__);
        if ( ssid_len < sizeof priv->ssid.ssid ) {
                memcpy(priv->ssid.ssid, ssid, ssid_len);
                priv->ssid.len = ssid_len;
        }
        else {
                return WL_INVALID_ARGS;
        }
        
        return WL_SUCCESS;
}

#else /* OWL_CONFIG_FW_AP */

int owl_sync_get_mode() {
        return IW_MODE_INFRA;
}

wl_err_t owl_sync_wlap_start(struct owl_net *priv,
                             char *ssid,
                             int ssid_len) {
        return WL_NOT_SUPPORTED;
}

wl_err_t owl_sync_set_channel(struct owl_net *priv,
                              uint8_t channel) {
        return WL_NOT_SUPPORTED;
}

wl_err_t owl_sync_wl_set_ssid(struct owl_net *priv, char* ssid, size_t ssid_len) {
        return WL_NOT_SUPPORTED;
}

#endif /* OWL_CONFIG_FW_AP */
